/*
 * File:   main.c
 *
 * PIC32MX250F128B for Ichigojam
 * Created on 2015/05/26, 14:40
 */

#include <stdio.h>
#include <stdlib.h>
#include <xc.h>
#include "font_for_pi.h"
//********** CONFIG **********************

#pragma config PMDL1WAY  = OFF
#pragma config IOL1WAY   = OFF
#pragma config FUSBIDIO  = OFF
#pragma config FVBUSONIO = OFF
#pragma config FPLLIDIV  = DIV_1
#pragma config FPLLMUL   = MUL_15
#pragma config FPLLODIV  = DIV_1
#pragma config FNOSC   = FRCPLL
#pragma config FSOSCEN   = OFF
#pragma config IESO      = OFF
#pragma config POSCMOD   = OFF
#pragma config OSCIOFNC  = OFF
#pragma config FPBDIV    = DIV_1
#pragma config FCKSM     = CSDCMD
#pragma config WDTPS     = PS1
#pragma config WINDIS    = OFF
#pragma config FWDTEN    = OFF
#pragma config FWDTWINSZ = WINSZ_50
#pragma config DEBUG     = OFF
#pragma config JTAGEN    = OFF
#pragma config ICESEL    = RESERVED
#pragma config PWP       = OFF
#pragma config BWP       = OFF
#pragma config CP        = OFF

//************ PROTOTYPES ******************

void ILI9325_INDEX(unsigned short);
void ILI9325_SET(unsigned short, unsigned short);
void ILI9325_DATA_WRITE(unsigned short);
void ILI9325_Initial(void);
void delay_ms(int);
void delay_sh(int);
void ILI9325_SETPIXEL(short, short, short);
void ILI9325_BOXFILL(short, short,short, short, short);
void ILI9325_BOX(short, short,short, short, short);
void capture(void);
void display(void);
void command(void);
void GR_PUTSTR(int ,int , char *, unsigned int ,unsigned );
void GR_PUTCH(int ,int ,unsigned int , unsigned int ,unsigned int );
void GR_PUTNUM(int ,int ,int , unsigned int ,unsigned int );
void GR_PUTHEX(int ,int ,int , unsigned int ,unsigned int );
void GR_SETPIXEL(short , short , short );

//************ DEFINE ****************
#define LCD_RST LATBbits.LATB14
#define VSYNC PORTBbits.RB5
#define VDATA PORTAbits.RA2
#define XSIZE 272
#define YSIZE 192

#define capline(x,y) { c0 = (VDATA)*16;\
                        asm("NOP");asm("NOP");asm("NOP");asm("NOP");asm("NOP");asm("NOP");asm("NOP");asm("NOP");\
                        frame_buffer[x][y] = (unsigned char)(c0 | VDATA);\
                        asm("NOP");asm("NOP");asm("NOP");\
                    }

//************ VALIABLES ******************
extern const unsigned char font8x8[];
unsigned char frame_buffer[XSIZE/2][YSIZE];
unsigned char attribute[34*2][24*2];
unsigned short colors[8]={
                            0x0000,0x001f,0x07e0,0x07ff,
                            0xf800,0xf81f,0xffe0,0xffff};

int works;
//************ MAIN ******************

int main() {
    int i,j,c,c0;

ANSELB = 0;
ANSELA = 0;
TRISB = 0x0000020;      // RB5 = SYNC input
TRISA = 0x0000004;      // RA2 = DATA input,RA4-RA3 is used as PMA1-0
ODCA = 0;
OSCCON = 0;

OSCTUN = 0;

IEC1CLR = 0x0004; // Disable PMP interrupt
PMCON = 0x0780;// Configure the addressing and polarities?
//PMMODE = 0x0251; // Configure the mode and timing
PMMODE = 0x0249; // Configure the mode and timing
PMADDR = 0x4000;    // CS1-2 ENABLE
PMAEN = 0x4003; // EnablChip Select lines, PMA0,PMA1
PMCON = 0x8780; // Stop PMP module and clear control register

ILI9325_Initial();
ILI9325_BOXFILL(10,10,120,120,0x00ff);
ILI9325_BOXFILL(90,200,229,309,0xff00);

//********* Attribute test *************
c0 = 0;
for (i=0;i<34*2;i++) {
    for (j=0;j<24*2;j++){
        attribute[i][j] = 0x07;
    }
}

 while(1) {
     capture();
     display();
     command();
 }
}
//************* ROUTINES *********************
void command(void)
{
    static unsigned char header = 0x55;
    unsigned char cmd[8];
    unsigned char c0,c;
    int i,j;
    
    for (i=0;i<8;i++){
        c = 0;
        for (j=0;j<4;j++) {
            c0=frame_buffer[(256-8)/2+j][i] & 0x11;
            c <<= 1;if ((c0>>4)!=0) c |= 1;
            c <<= 1;if ((c0&1)!=0) c |= 1;
        }
        cmd[i]=c;
        //GR_PUTHEX(5+36*i,0,cmd[i],0,0xffff);
    }
    if (cmd[0]==header){
        header = header ^ 0xff;
        if (cmd[1]==0xc0){                      //**** TEXT ATTRIBUTE SET *****
            for (i=cmd[2]*2;i<=cmd[4]*2;i++){
                for (j=cmd[3]*2;j<=cmd[5]*2;j++) {
                    attribute[i][j]=cmd[6];
                    attribute[i][j+1]=cmd[6];
                    attribute[i+1][j]=cmd[6];
                    attribute[i+1][j+1]=cmd[6];
                }
            }
        }
        if (cmd[1]==0xc1){                      //**** SEMI-GRAPGIC ATTRIBUTE SET *****
            for (i=cmd[2];i<=cmd[4];i++){
                for (j=cmd[3];j<=cmd[5];j++) {
                    attribute[i][j]=cmd[6];
                }
            }
        }
    }

}


void capture(void)
{
    int doing=1;
    int i,k,c0;
    int state = 0;
    int t0;

    t0=0;
    while(doing==1){
        LATBbits.LATB4 = ~LATBbits.LATB4;
        switch (state) {
           case 0:
                    while(VSYNC == 1){};
                    t0 = 0;
                    state = 1;
                    break;
            case 1:
                  // **** Wait for VBLANK END *****
                    while(VSYNC == 0){t0++;}
                    if (t0<=100){
                        state=0;t0=0;
                    }else{
                        while(VSYNC == 1){};
                        state =2;t0=0;
                    }
                    break;
            case 2:
                  // **** Wait for VBLANK END *****
                    while(VSYNC == 0){t0++;}
                    if (t0<=100){
                        state=3;t0=0;
                    }else{
                        while(VSYNC == 1){};
                        t0=0;
                    }
                    break;
            case 3:
                  //LATBbits.LATB5 = 1;
                  // **** Wait for VBLANK END *****
                  for (i=0;i<28;i++) {
                       while(VSYNC == 1){};
                       while(VSYNC == 0){};
                  }
                  while(VSYNC == 1){};
                  for (i=0;i<192;i++){
                       while(VSYNC == 0){};
                       for(k=0;k<280;k++){};                     // ** Wait for data timing **
                       asm("NOP");asm("NOP");asm("NOP");asm("NOP");asm("NOP");asm("NOP");asm("NOP");asm("NOP");asm("NOP");asm("NOP");asm("NOP");
                       asm("NOP");asm("NOP");asm("NOP");asm("NOP");
                       capline(0,i);capline(1,i);capline(2,i);capline(3,i);capline(4,i);capline(5,i);capline(6,i);capline(7,i);capline(8,i);capline(9,i);
                       capline(10,i);capline(11,i);capline(12,i);
                       capline(13,i);capline(14,i);capline(15,i);capline(16,i);capline(17,i);capline(18,i);capline(19,i);
                       capline(20,i);capline(21,i);capline(22,i);capline(23,i);capline(24,i);capline(25,i);capline(26,i);capline(27,i);capline(28,i);capline(29,i);
                       capline(30,i);capline(31,i);capline(32,i);capline(33,i);capline(34,i);capline(35,i);capline(36,i);capline(37,i);capline(38,i);capline(39,i);
                       capline(40,i);capline(41,i);capline(42,i);capline(43,i);capline(44,i);capline(45,i);capline(46,i);capline(47,i);capline(48,i);capline(49,i);
                       asm("NOP");
                       capline(50,i);capline(51,i);capline(52,i);capline(53,i);capline(54,i);capline(55,i);capline(56,i);capline(57,i);capline(58,i);capline(59,i);
                       capline(60,i);capline(61,i);capline(62,i);capline(63,i);capline(64,i);capline(65,i);capline(66,i);capline(67,i);capline(68,i);capline(69,i);
                       capline(70,i);capline(71,i);capline(72,i);capline(73,i);capline(74,i);capline(75,i);capline(76,i);capline(77,i);capline(78,i);capline(79,i);
                       capline(80,i);capline(81,i);capline(82,i);capline(83,i);capline(84,i);capline(85,i);capline(86,i);capline(87,i);capline(88,i);capline(89,i);
                       capline(90,i);capline(91,i);capline(92,i);capline(93,i);capline(94,i);capline(95,i);capline(96,i);capline(97,i);capline(98,i);capline(99,i);
                       asm("NOP");
                       capline(100,i);capline(101,i);capline(102,i);capline(103,i);capline(104,i);capline(105,i);capline(106,i);capline(107,i);capline(108,i);capline(109,i);
                       capline(110,i);capline(111,i);capline(112,i);capline(113,i);capline(114,i);capline(115,i);capline(116,i);capline(117,i);capline(118,i);capline(119,i);
                       capline(120,i);capline(121,i);capline(122,i);capline(123,i);capline(124,i);capline(125,i);capline(126,i);capline(127,i);capline(128,i);capline(129,i);
                       capline(130,i);capline(131,i);capline(132,i);capline(133,i);capline(134,i);capline(135,i);
                       if (i<191) { while(VSYNC == 1){}};
                  }
                  state = 0;
                  doing = 0;
                  break;
            default:
                state = 0;
                break;

        }
    }
}

void display(void)
{
    static int BLINK = 0;
    int H,V;
    unsigned short c0,data,bc,fc;
    unsigned char attr;

    for (H=0;H<192;H++){
        for (V=0;V<272;V+=2){
            if (V==0) {
                ILI9325_SET(0x0020,H+24);
                ILI9325_SET(0x0021,320-24);
                ILI9325_INDEX(0x0022);
            }
            if ((V%4)==0) {
                attr = attribute[(V/4)][H/4];
                bc = colors[(attr>>4)%8];
                fc = colors[attr%8];
            }
            c0 = frame_buffer[V/2][H] & 0x11;
            if ((c0>>4) !=0) {data = fc;}else{data=bc;}
            if (((attr & 0x80)!=0)&&((H%8)==7)){data=fc;}         // **** UNDERLINE *********
            if (((attr & 0x08)!=0)&&(BLINK>10)){data=bc;}         // **** BLINK *********
            while((PMSTAT & 0x8000)!=0){};PMADDR = 0x4003;PMDIN = (unsigned char)(data >> 8);
            while((PMSTAT & 0x8000)!=0){};PMDIN = (unsigned char)(data & 0x00ff);
            if ((c0<<4) !=0) {data = fc;}else{data=bc;}
            if (((attr & 0x80)!=0)&&((H%8)==7)){data=fc;}         // **** UNDERLINE *********
            if (((attr & 0x08)!=0)&&(BLINK>10)){data=bc;}         // **** BLINK *********
            while((PMSTAT & 0x8000)!=0){};PMDIN = (unsigned char)(data >> 8);
            while((PMSTAT & 0x8000)!=0){};PMDIN = (unsigned char)(data & 0x00ff);
        }
    }
    BLINK++;if (BLINK>20) BLINK=0;
}

void ILI9325_BOXFILL(short x0, short y0,short x1, short y1, short rgb)
{
    short x,y;
    for (x=x0;x<=x1;x++){
        ILI9325_SETPIXEL(x,y0,rgb);
        for (y=y0+1;y<=y1;y++){
            ILI9325_SETPIXEL(-1,0,rgb);
        }
    }
}
void ILI9325_BOX(short x0, short y0,short x1, short y1, short rgb)
{
    short x,y;
    ILI9325_SETPIXEL(x1,y0,rgb);
    for (x=x0+1;x<=x1;x++){
        ILI9325_SETPIXEL(-1,0,rgb);
    }
    ILI9325_SETPIXEL(x1,y1,rgb);
    for (x=x0+1;x<=x1;x++){
        ILI9325_SETPIXEL(-1,0,rgb);
    }
    for (y=y0;y<=y1;y++){
        for (x=x0;x<=x1;x++){
            ILI9325_SETPIXEL(x,y0,rgb);
            ILI9325_SETPIXEL(x,y1,rgb);
        }
    }
}

void ILI9325_SETPIXEL(short x, short y, short rgb)
{
    if (x>=0) {
     ILI9325_SET(0x0020,x);
     ILI9325_SET(0x0021,320-y);
     ILI9325_INDEX(0x0022);
    }
    ILI9325_DATA_WRITE(rgb);
}

void ILI9325_INDEX(unsigned short index)
{
 delay_sh(1);
 while((PMSTAT & 0x8000)!=0){};PMADDR = 0x4000;PMDIN = (unsigned char)(index >> 8);
 delay_sh(1);
 while((PMSTAT & 0x8000)!=0){};PMDIN = (unsigned char)(index & 0x00ff);
 while((PMSTAT & 0x8000)!=0){};PMADDR = 0x4003;
}

void ILI9325_SET(unsigned short index, unsigned short data)
{
 delay_sh(1);
 while((PMSTAT & 0x8000)!=0){};PMADDR = 0x4000;PMDIN = (unsigned char)(index >> 8);
 delay_sh(1);
 while((PMSTAT & 0x8000)!=0){};PMDIN = (unsigned char)(index & 0x00ff);
 while((PMSTAT & 0x8000)!=0){};PMADDR = 0x4003;PMDIN = (unsigned char)(data >> 8);
 while((PMSTAT & 0x8000)!=0){};PMDIN = (unsigned char)(data & 0x00ff);
}

void ILI9325_DATA_WRITE(unsigned short data)
{
 while((PMSTAT & 0x8000)!=0){};PMADDR = 0x4003;PMDIN = (unsigned char)(data >> 8);
 while((PMSTAT & 0x8000)!=0){};PMADDR = 0x4003;PMDIN = (unsigned char)(data & 0x00ff);
}

void ILI9325_Initial(void)
{

 LCD_RST = 0; // **** LCD RESET **********
 delay_ms(200);
 LCD_RST = 1;

 //******* SET REGISTERS **********
 ILI9325_SET(0x0001,0x0100);
 ILI9325_SET(0x0002,0x0700);
 ILI9325_SET(0x0003,0x1008);
 ILI9325_SET(0x0004,0x0000);
 ILI9325_SET(0x0008,0x0207);
 ILI9325_SET(0x0009,0x0000);
 ILI9325_SET(0x000a,0x0000);
 ILI9325_SET(0x000c,0x0000);
 ILI9325_SET(0x000d,0x0000);
 ILI9325_SET(0x000f,0x0000);

 ILI9325_SET(0x0010,0x0000);
 ILI9325_SET(0x0011,0x0000);
 ILI9325_SET(0x0012,0x0000);
 ILI9325_SET(0x0013,0x0000);
 delay_ms(200);
 ILI9325_SET(0x0010,0x1490);
 ILI9325_SET(0x0011,0x0227);
 delay_ms(200);
 ILI9325_SET(0x0012,0x001a);
 delay_ms(200);
 ILI9325_SET(0x0013,0x1400);
 ILI9325_SET(0x0029,0x0019);
 ILI9325_SET(0x002b,0x000b);
 delay_ms(200);
 ILI9325_SET(0x0020,0x0000);
 ILI9325_SET(0x0021,0x0000);
 ILI9325_SET(0x0030,0x0000);
 ILI9325_SET(0x0031,0x0607);
 ILI9325_SET(0x0032,0x0305);
 ILI9325_SET(0x0035,0x0000);
 ILI9325_SET(0x0036,0x1604);
 ILI9325_SET(0x0037,0x0204);
 ILI9325_SET(0x0038,0x0001);
 ILI9325_SET(0x0039,0x0707);
 ILI9325_SET(0x003c,0x0000);
 ILI9325_SET(0x003d,0x000f);
 ILI9325_SET(0x0050,0x0000);
 ILI9325_SET(0x0051,0x00ef);
 ILI9325_SET(0x0052,0x0000);
 ILI9325_SET(0x0053,0x013f);
 ILI9325_SET(0x0060,0xa700);
 ILI9325_SET(0x0061,0x0001);
 ILI9325_SET(0x006a,0x0000);
 ILI9325_SET(0x0080,0x0000);
 ILI9325_SET(0x0081,0x0000);
 ILI9325_SET(0x0082,0x0000);
 ILI9325_SET(0x0083,0x0000);
 ILI9325_SET(0x0084,0x0000);
 ILI9325_SET(0x0085,0x0000);
 ILI9325_SET(0x0090,0x0010);
 ILI9325_SET(0x0092,0x0600);
 ILI9325_SET(0x0093,0x0003);
 ILI9325_SET(0x0095,0x0110);
 ILI9325_SET(0x0097,0x0000);
 ILI9325_SET(0x0098,0x0000);
 ILI9325_SET(0x0007,0x0133);
  delay_ms(200);
}

void delay_ms(int tim)
{
 int t,t1;
    for (t=0;t<tim;t++){
         for (t1=0;t1<30000;t1++) {
             works++;
        }
    }
 }

void delay_sh(int tim)
{
 int t,t1;
    for (t=0;t<tim;t++){
         for (t1=0;t1<1;t1++) {
             works++;
        }
    }
 }

//********** Line *************

void GR_LINE(int x0,int y0,int x1,int y1,unsigned int col)
{
  int i,dx,dy,e,sx,sy,xx0,yy0;

  xx0=x0;
  yy0=y0;

  if (x1 > x0) {
		dx = x1-x0;
		sx = 1;
  }else{
		dx = x0-x1;
		sx = -1;
  }

  if (y1 > y0) {
		dy = y1-y0;
		sy = 1;
  }else{
		dy = y0-y1;
		sy = -1;
  }

  if ( dx > dy ) {
    e = -dx;
    for ( i = 0 ; i <= dx ; i++ ) {
      GR_SETPIXEL(xx0,yy0,col);
      xx0 += sx;
      e += 2 * dy;
      if ( e >= 0 ) {
        yy0 += sy;
        e -= 2 * dx;
      }
    }
  } else {
    e = -dy;
    for ( i = 0 ; i <= dy ; i++ ) {
      GR_SETPIXEL(xx0,yy0,col);
      yy0 += sy;
      e += 2 * dx;
      if ( e >= 0 ) {
        xx0 += sx;
        e -= 2 * dy;
      }
    }
  }
}

//********** Put string *************
void GR_PUTSTR(int x,int y, char *dt, unsigned int col,unsigned bg) {
    char *p0;
	int x0;

	p0 = dt;
	x0 = x;

	while (*p0 != 0x00) {
		GR_PUTCH(x0,y,*p0++,col,bg);
		x0 += 8;
	}
}
void GR_PUTCH(int x,int y,unsigned int dt, unsigned int col,unsigned int bg) {

	int p0,i,j,k,l;
	unsigned int c0;
	unsigned int cl;

	p0 = (dt - 0x20) * 8;

	for (i=0;i<8;i++) {
		c0 = font8x8[p0+i];
		for (j=0;j<8;j++) {
			if ((c0 & 0x01) != 0) {
				cl = col;
			}else{
				cl = bg;
			}
                        GR_SETPIXEL(x+i,y+j,cl);
                        c0 >>= 1;
                   }
	}
}
//********** Put Number *************
void GR_PUTNUM(int x,int y,int v, unsigned int col,unsigned int bg) {

	int i,x0,vv;

	x0 = x+3*8;
	vv=v;

	for (i=0;i<4;i++) {
		if ((i==0)||(vv>=0)) {
			GR_PUTCH(x0,y,0x30+(vv%10),col,bg);
		}
		x0 -= 8;
		vv /=10;
	}
}
//********** Put Number *************
void GR_PUTHEX(int x,int y,int v, unsigned int col,unsigned int bg) {

	int i,x0,vv;
	x0 = x+8*3;
	vv=v;
	for (i=0;i<4;i++) {
              if ((vv%16)<10){
        	GR_PUTCH(x0,y,0x30+(vv%16),col,bg);
              }else{
                GR_PUTCH(x0,y,0x41+((vv%16)-10),col,bg);
              }
            x0 -= 8;
            vv /=16;
	}
}

void GR_SETPIXEL(short x, short y, short rgb)
{
   ILI9325_SET(0x0020,y);
   ILI9325_SET(0x0021,320-x);
   ILI9325_INDEX(0x0022);
  ILI9325_DATA_WRITE(rgb);
}
